"""
Package for modeling epidemics using the SEIR model
"""
__author__ = 'Janne Solanpää'
__author_email__ = 'janne@solanpaa.fi'
__copyright__ = '© Janne Solanpää'
__credits__ = ''
__license__ = 'MIT'
__version__ = '0.0.1'
__mainainer__ = 'Janne Solanpää'
__status__ = 'Alpha'
__name__ = 'seir'
